import { request as httpRequest } from "http";
import { request as httpsRequest } from "https";
import { URL } from "url";
import { dispose, AbstractDisposable, toErrorHandler, addTeardown, bindDisposables, addDisposable, } from "@reactive-js/core/lib/disposable.mjs";
import { defer, pipe, returns } from "@reactive-js/core/lib/functions.mjs";
import { createWritableIOSink, createDisposableNodeStream, createReadableIOSource, } from "@reactive-js/core/lib/node.mjs";
import { dispatch, subscribe, scan, onNotify, map, keepType, using, createSubject, merge, dispatchTo, } from "@reactive-js/core/lib/observable.mjs";
import { none, isSome } from "@reactive-js/core/lib/option.mjs";
import { stream } from "@reactive-js/core/lib/streamable.mjs";
import { httpRequestToUntypedHeaders, parseHttpResponseFromHeaders, } from "../../http.mjs";
class ResponseBody extends AbstractDisposable {
    constructor(resp) {
        super();
        this.resp = resp;
        this.consumed = false;
        addTeardown(this, _ => {
            resp.removeAllListeners();
            resp.destroy();
        });
        const onError = toErrorHandler(this);
        resp.on("error", onError);
    }
    stream(scheduler, replayCount) {
        if (this.consumed) {
            throw new Error("Response body already consumed");
        }
        this.consumed = true;
        const responseStream = stream(createReadableIOSource(defer(this.resp, createDisposableNodeStream)), scheduler, replayCount);
        bindDisposables(this, responseStream);
        return responseStream;
    }
}
export const createHttpClient = (options = {}) => request => using((scheduler) => {
    const { method, uri } = request;
    const url = uri instanceof URL ? uri : new URL(uri.toString());
    const headers = httpRequestToUntypedHeaders(request);
    const nodeRequestOptions = {
        ...options,
        headers,
        method,
    };
    const req = uri.protocol === "https:"
        ? httpsRequest(url, nodeRequestOptions)
        : uri.protocol === "http:"
            ? httpRequest(url, nodeRequestOptions)
            : (() => {
                throw new Error();
            })();
    const requestSink = stream(createWritableIOSink(defer(req, createDisposableNodeStream)), scheduler);
    const requestBody = stream(request.body, scheduler);
    const onContinue = () => {
        const reqSubscription = pipe(requestSink, onNotify(dispatchTo(requestBody)), subscribe(scheduler));
        const dataSubscription = pipe(requestBody, onNotify(dispatchTo(requestSink)), subscribe(scheduler));
        addDisposable(dataSubscription, reqSubscription),
            addDisposable(requestBody, dataSubscription);
        addDisposable(requestBody, requestSink);
    };
    if (request.expectContinue) {
        req.on("continue", onContinue);
    }
    else {
        onContinue();
    }
    const responseSubject = createSubject();
    const onResponse = (resp) => {
        var _a;
        dispose(requestBody);
        const body = new ResponseBody(resp);
        bindDisposables(responseSubject, body);
        const response = parseHttpResponseFromHeaders((_a = resp.statusCode) !== null && _a !== void 0 ? _a : -1, resp.headers, body);
        dispatch(responseSubject, response);
    };
    req.on("response", onResponse);
    return [requestBody, responseSubject];
}, (requestBody, responseSubject) => merge(pipe(requestBody, scan(([increment, count], ev) => ev.type === 1
    ? [ev.data.length, count + increment]
    : [-1, count + increment], returns([0, 0])), map(([increment, count]) => increment < 0
    ? { type: 3 }
    : count > 0
        ? {
            type: 2,
            count,
        }
        : none), keepType(isSome)), pipe(responseSubject, map(response => ({
    type: 4,
    response,
})))));
